@@ -1,24 +1,64 @@
 /* **********************************************************
  * Copyright (C) 1998-2000 VMware, Inc.
  * All Rights Reserved
- * $Id: x86.h,v 1.4 2003/02/16 10:09:57 bad Exp $
+ * $Id: x86.h,v 1.5 2003/02/16 15:32:34 bad Exp $
  * **********************************************************/
 
 #ifndef _X86_H_
 #define _X86_H_
 
+#define INCLUDE_ALLOW_USERLEVEL
+#define INCLUDE_ALLOW_MONITOR
+#define INCLUDE_ALLOW_MODULE
+#define INCLUDE_ALLOW_VMKERNEL
+#include "includeCheck.h"
+
 #include "vm_types.h"
+#include "x86apic.h"
+#include "x86desc.h"
 
-
-#define SIZE_8BIT  1
-#define SIZE_16BIT 2
-#define SIZE_24BIT 3
-#define SIZE_32BIT 4
-#define SIZE_48BIT 6
-#define SIZE_64BIT 8
-#define SIZE_80BIT 10
+#define SIZE_8BIT   1
+#define SIZE_16BIT  2
+#define SIZE_24BIT  3
+#define SIZE_32BIT  4
+#define SIZE_48BIT  6
+#define SIZE_64BIT  8
+#define SIZE_80BIT  10
 #define SIZE_128BIT 16
+#define SIZE_M512   512
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * OpsizeValue --
+ *
+ *	Called by code that can handle 32, 16, or 8 bits values to return
+ *	only the defined bits for the current size.
+ *
+ * Results:
+ *      Value of the specified bits, zero-extended out to 32 bits.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+#if defined(LOWORD) && defined(LOBYTE) // Windows doesn't have LOWORD here -- edward
+static INLINE uint32
+OpsizeValue(uint32 value,	// IN: The 32 bit value to be sized.
+	    int opSize)		// IN: the size 
+{
+#ifdef ASSERT
+   ASSERT(opSize == SIZE_8BIT || opSize == SIZE_16BIT || opSize == SIZE_32BIT);
+#endif
+
+   return opSize == SIZE_8BIT  ? LOBYTE(value) : 
+         (opSize == SIZE_16BIT ? LOWORD(value) : value);
+}
+#endif
+
+#define NUM_IDT_VECTORS 256
 
 
 /*
@@ -66,7 +106,7 @@
 #define CR4_OSXMMEXCPT 0x00000400 /* #XF exception enable PIII only */
 
 #define CR4_RESERVED   0xfffff800
-#define CR4_NOT_IMP    0x00000026
+#define CR4_NOT_IMP    0x00000006
 
 #define CR4_RESERVED_WO_KNI   (CR4_RESERVED|CR4_OSXMMEXCPT)
 #define CR4_NOT_IMP_WO_KNI    (CR4_NOT_IMP|CR4_OSXMMEXCPT)
@@ -141,48 +181,120 @@
 
 
 
-/* www.sandpile.org/80x86/cpuid.shtml */
-#define CPUID_FEATURE_FPU    0x00000001
-#define CPUID_FEATURE_VME    0x00000002
-#define CPUID_FEATURE_DBGE   0x00000004
-#define CPUID_FEATURE_PGSZE  0X00000008
-#define CPUID_FEATURE_TSC    0x00000010
-#define CPUID_FEATURE_MSR    0x00000020
-#define CPUID_FEATURE_PAE    0x00000040
-#define CPUID_FEATURE_MCK    0x00000080
-#define CPUID_FEATURE_CPMX   0x00000100
-#define CPUID_FEATURE_APIC   0x00000200
-#define CPUID_FEATURE_RSVD1  0x00000400
-#define CPUID_FEATURE_SEP    0x00000800
-#define CPUID_FEATURE_MTTR   0x00001000
-#define CPUID_FEATURE_PGE    0x00002000 
-#define CPUID_FEATURE_MCA    0x00004000
-#define CPUID_FEATURE_CMOV   0x00008000 
-#define CPUID_FEATURE_PAT    0x00010000
-#define CPUID_FEATURE_36PG   0x00020000
-#define CPUID_FEATURE_RSVD2  0x00040000 /* pentium II, mendocino? */
-#define CPUID_FEATURE_RSVD3  0x00080000
-#define CPUID_FEATURE_RSVD4  0x00100000
-#define CPUID_FEATURE_RSVD5  0x00200000
-#define CPUID_FEATURE_RSVD6  0x00400000
-#define CPUID_FEATURE_MMX    0x00800000 
-#define CPUID_FEATURE_FXSAVE 0x01000000
-#define CPUID_FEATURE_XMM    0x02000000  /* pentium III */
-#define CPUID_FEATURE_RSVD   0xfc780000 
+/* http://www.sandpile.org/ia32/cpuid.htm */
+
+/* Common CPUID feature bits */
+#define CPUID_FEATURE_COMMON_FPU    0x00000001
+#define CPUID_FEATURE_COMMON_VME    0x00000002
+#define CPUID_FEATURE_COMMON_DBGE   0x00000004
+#define CPUID_FEATURE_COMMON_PGSZE  0X00000008
+#define CPUID_FEATURE_COMMON_TSC    0x00000010
+#define CPUID_FEATURE_COMMON_MSR    0x00000020
+#define CPUID_FEATURE_COMMON_PAE    0x00000040
+#define CPUID_FEATURE_COMMON_MCK    0x00000080
+#define CPUID_FEATURE_COMMON_CPMX   0x00000100
+#define CPUID_FEATURE_COMMON_APIC   0x00000200
+#define CPUID_FEATURE_COMMON_RSVD10 0x00000400
+#define CPUID_FEATURE_COMMON_SEP    0x00000800
+#define CPUID_FEATURE_COMMON_MTTR   0x00001000
+#define CPUID_FEATURE_COMMON_PGE    0x00002000 
+#define CPUID_FEATURE_COMMON_MCA    0x00004000
+#define CPUID_FEATURE_COMMON_CMOV   0x00008000 
+#define CPUID_FEATURE_COMMON_PAT    0x00010000
+#define CPUID_FEATURE_COMMON_36PG   0x00020000
+/* feature bits 18-22 */
+#define CPUID_FEATURE_COMMON_MMX    0x00800000 
+#define CPUID_FEATURE_COMMON_FXSAVE 0X01000000
+#define CPUID_FEATURE_COMMON_XMM    0x02000000
+/* feature bits 26-31 */
+
+/* Intel CPUID feature bits */
+#define CPUID_FEATURE_INTEL_PSN    0x00040000
+#define CPUID_FEATURE_INTEL_CLFL   0x00080000
+#define CPUID_FEATURE_INTEL_RSVD20 0x00100000
+#define CPUID_FEATURE_INTEL_DTES   0x00200000
+#define CPUID_FEATURE_INTEL_ACPI   0x00400000
+/* CPUID_FEATURE_COMMON_MMX */
+/* CPUID_FEATURE_COMMON_FXSAVE */
+/* CPUID_FEATURE_COMMON_XMM    */
+#define CPUID_FEATURE_INTEL_XMM2   0x04000000
+#define CPUID_FEATURE_INTEL_SS     0x08000000 
+#define CPUID_FEATURE_INTEL_RSVD28 0x10000000  // Pentium 4 sets it to 1!
+#define CPUID_FEATURE_INTEL_TM     0x20000000 
+#define CPUID_FEATURE_INTEL_RSVD30 0x40000000 
+#define CPUID_FEATURE_INTEL_RSVD31 0x80000000 
+
+#define CPUID_FEATURE_INTEL_RSVD   0xc0100400 
+
+/* AMD CPUID feature bits */
+#define CPUID_FEATURE_AMD_RSVD18 0X00040000
+#define CPUID_FEATURE_AMD_RSVD19 0X00080000
+#define CPUID_FEATURE_AMD_RSVD20 0X00100000
+#define CPUID_FEATURE_AMD_RSVD21 0X00200000
+#define CPUID_FEATURE_AMD_MMX1   0X00400000
+/* CPUID_FEATURE_COMMON_MMX */
+/* CPUID_FEATURE_COMMON_FXSAVE */
+/* CPUID_FEATURE_COMMON_XMM    */
+#define CPUID_FEATURE_AMD_RSVD26 0X04000000
+#define CPUID_FEATURE_AMD_RSVD27 0X08000000
+#define CPUID_FEATURE_AMD_RSVD28 0X10000000
+#define CPUID_FEATURE_AMD_EXTENDED_LM     0X20000000
+#define CPUID_FEATURE_AMD_EXTENDED_3DNOW1 0X40000000
+#define CPUID_FEATURE_AMD_EXTENDED_3DNOW  0X80000000
+
+#define CPUID_FEATURE_AMD_RSVD   0x1c3c0400 
+
+#define CPUID_STEPPING(_eax) ((_eax) & 0xf)
 
 #define CPUID_FAMILY(_eax)   (((_eax) >> 8) & 0xf)
 
+/* Intel CPU Family */
 #define CPUID_FAMILY_486     4
 #define CPUID_FAMILY_P5      5
 #define CPUID_FAMILY_P6      6
+#define CPUID_FAMILY_EXTENDED 15
+
+/* AMD CPU Family */
+#define CPUID_FAMILY_5x86    4
+#define CPUID_FAMILY_K5      5
+#define CPUID_FAMILY_K6      5
+#define CPUID_FAMILY_K7      6
+
+#define CPUID_EXTENDED_FAMILY(_eax)   (((_eax) >> 20) & 0xff)
+
+#define CPUID_EXTENDED_FAMILY_PENTIUM4	0
+
+#define CPUID_FAMILY_IS_486(_eax) (CPUID_FAMILY(_eax) == CPUID_FAMILY_486)
+#define CPUID_FAMILY_IS_P5(_eax)  (CPUID_FAMILY(_eax) == CPUID_FAMILY_P5)
+#define CPUID_FAMILY_IS_P6(_eax)  (CPUID_FAMILY(_eax) == CPUID_FAMILY_P6)
+#define CPUID_FAMILY_IS_PENTIUM4(_eax)  \
+           (CPUID_FAMILY(_eax) == CPUID_FAMILY_EXTENDED && \
+            CPUID_EXTENDED_FAMILY(_eax) == CPUID_EXTENDED_FAMILY_PENTIUM4)
+
+
+#define CPUID_MODEL(_eax)    (((_eax) >> 4) & 0xf)
+
+#define CPUID_MODEL_PPRO       1
+#define CPUID_MODEL_PII_3      3
+#define CPUID_MODEL_PII_5      5
+#define CPUID_MODEL_CELERON_6  6
+#define CPUID_MODEL_PIII_7     7
+#define CPUID_MODEL_PIII_8     8
+#define CPUID_MODEL_PIII_A     10
+#define CPUID_MODEL_EXTENDED   15
+
+#define CPUID_EXTENDED_MODEL(_eax)    (((_eax) >> 16) & 0xf)
 
-#ifndef MSR_TSC
 #define MSR_TSC              0x00000010
-#endif /* MSR_TSC */
+#define MSR_PERFCTR0         0x000000c1
+#define MSR_PERFCTR1         0x000000c2
 #define MSR_MTRR_CAP         0x000000fe
+#define MSR_L2CFG            0x0000011e
 #define MSR_SYSENTER_CS      0x00000174
 #define MSR_SYSENTER_ESP     0x00000175
 #define MSR_SYSENTER_EIP     0x00000176
+#define MSR_EVNTSEL0         0x00000186
+#define MSR_EVNTSEL1         0x00000187
 
 
 typedef uint32 CReg;
@@ -203,11 +315,11 @@ typedef uint16 Selector;
 #define SELECTOR_INDEX_SHIFT     3
 #define SELECTOR_INDEX_MASK      0xfff8
 
-#define SELECTOR_RPL(_sel)       ((_sel) & SELECTOR_RPL_MASK)
-#define SELECTOR_TABLE(_sel)     (((_sel) & SELECTOR_TI_MASK) >> SELECTOR_TI_SHIFT)
-#define SELECTOR_INDEX(_sel)     ((_sel) >> SELECTOR_INDEX_SHIFT)
-
-#define NULL_SELECTOR(_sel)      (!((_sel) & ~SELECTOR_RPL_MASK))
+#define SELECTOR_RPL(_sel)       (((Selector)(_sel)) & SELECTOR_RPL_MASK)
+#define SELECTOR_TABLE(_sel)     ((((Selector)(_sel)) & SELECTOR_TI_MASK) >> SELECTOR_TI_SHIFT)
+#define SELECTOR_INDEX(_sel)     (((Selector)(_sel)) >> SELECTOR_INDEX_SHIFT)
+#define NULL_SELECTOR(_sel)      (!(((Selector)(_sel)) & ~SELECTOR_RPL_MASK))
+#define SELECTOR_CLEAR_RPL(_sel) ((_sel) & ~SELECTOR_RPL_MASK)
 
 /*
  *   tasks
@@ -283,201 +395,14 @@ typedef struct {
 
 
 /*
- *   segment descriptors
+ *   far pointers
  */
 
-typedef struct Descriptor {
-   unsigned   limit_lo  : 16;
-   unsigned   base_lo   : 16;
-   unsigned   base_mid  : 8;
-   unsigned   type      : 4;
-   unsigned   S         : 1;
-   unsigned   DPL       : 2;
-   unsigned   present   : 1;
-   unsigned   limit_hi  : 4;
-   unsigned   AVL       : 1;
-   unsigned   unused    : 1;
-   unsigned   DB        : 1;
-   unsigned   gran      : 1;
-   unsigned   base_hi   : 8;
-} Descriptor;
-
-typedef union {
-   Descriptor desc;
-   uint32     word[2];
-} DescriptorUnion;
-
-
-#define EXPANDEDLIMIT(_desc)((_desc).gran? (LIMIT(_desc) << 12) | 0xfff : LIMIT(_desc))
-#define LIMIT(_desc)        (((_desc).limit_hi<<16) | (_desc).limit_lo)
-#define BASE(_desc)         (((_desc).base_hi<<24) | ((_desc).base_mid<<16) | (_desc).base_lo)
-
-#define LIMIT_LO(_dw)       (((int)(_dw)) & 0xffff)
-#define LIMIT_HI(_dw)       ((((int)(_dw)) >> 16) & 0xf)
-
-#define BASE_LO(_dw)        (((int)(_dw)) & 0xffff)
-#define BASE_MID(_dw)       ((((int)(_dw)) >> 16) & 0xff)
-#define BASE_HI(_dw)        ((((int)(_dw)) >> 24) & 0xff)
-
-#define DT_CODE(_desc)               ( ((_desc).S) && (((_desc).type & 0x8) == 0x8))
-#define DT_CONFORMING_CODE(_desc)    ( ((_desc).S) && (((_desc).type & 0xc) == 0xc))
-#define DT_NONCONFORMING_CODE(_desc) ( ((_desc).S) && (((_desc).type & 0xc) == 0x8))
-#define DT_READABLE_CODE(_desc)      ( ((_desc).S) && (((_desc).type & 0xa) == 0xa))
-#define DT_DATA(_desc)               ( ((_desc).S) && (((_desc).type & 0x8) == 0x0))
-#define DT_WRITEABLE_DATA(_desc)     ( ((_desc).S) && (((_desc).type & 0xa) == 0x2))
-#define DT_EXPAND_DOWN(_desc)        ( ((_desc).S) && (((_desc).type & 0xc) == 0x4))
-#define DT_CALL_GATE(_desc)          (!((_desc).S) && (((_desc).type & 0x7) == 0x4))
-#define DT_LDT(_desc)                (!((_desc).S) && (((_desc).type & 0xf) == 0x2))
-#define DT_TASK_GATE(_desc)          (!((_desc).S) && (((_desc).type & 0xf) == 0x5))
-#define DT_TSS(_desc)                (!((_desc).S) && (((_desc).type & 0x5) == 0x1))
-#define DT_AVAIL_TSS(_desc)          (!((_desc).S) && (((_desc).type & 0x7) == 0x1))
-
-#define DT_ACCESS                    0x2
-#define DT_32BIT                     0x8
-#define DT_TSS_BUSY                  0x2
-
-#define GT_TASK(_gate)               (((_gate).type & 0x1f) == 0x05)
-#define GT_INTR(_gate)               (((_gate).type & 0x17) == 0x06)
-#define GT_TRAP(_gate)               (((_gate).type & 0x17) == 0x07)
-#define GT_32BIT                     0x08
-
-#define DATA_DESC           0x2
-#define CODE_DESC           0xa
-#define TASK_DESC           0x9  /* TTS available */
-#define TASK_DESC_BUSY      0xb  /* TSS busy */
-
-
-/*
- *----------------------------------------------------------------------
- *
- * Desc_Limit --
- *
- *      Return the limit in bytes of the descriptor.
- *
- * Results:
- *      The limit in bytes of the descriptor.
- *
- * Side effects:
- *      None.
- *
- *----------------------------------------------------------------------
- */
-
-static INLINE uint32
-Desc_Limit(Descriptor *descp)
-{
-   return (descp->gran) ? (LIMIT(*descp) << 12) | 0xfff : LIMIT(*descp);
-}
-
-
-/*
- *----------------------------------------------------------------------
- *
- * Desc_InBounds and Desc_LegalAccess --
- *
- *      Check whether the access is within the bounds of the segment.
- *
- * Results:
- *      TRUE is the access is within the bouns of the segment, FALSE
- *      otherwise.
- *
- * Side effects:
- *      None.
- *
- * Bugs:
- *      None.
- *
- *----------------------------------------------------------------------
- */
-
-static INLINE Bool
-Desc_InBounds(VA vaddr, VA limit, uint32 size, Bool expandDown, unsigned DB) {
-   if (expandDown) {
-      VA end = DB ? 0xffffffff : 0xffff; 
-      return  ((vaddr >= limit) && (vaddr <= end) && ((size-1) <= (end - vaddr)));
-   } else {
-      return  ((vaddr <= limit) && ((size-1) <= (limit - vaddr)));
-   }
-}
-
-static INLINE Bool
-Desc_LegalAccess(Descriptor *descp, VA vaddr, uint32 size)
-{
-   return Desc_InBounds(vaddr,Desc_Limit(descp),
-                        size,DT_EXPAND_DOWN(*descp),descp->DB);
-}
-
-
-/*
- *   call gates
- */
-
-typedef struct {
-   unsigned   offset_lo : 16;
-   unsigned   segment   : 16;
-   unsigned   params    : 5;
-   unsigned   unused    : 3;
-   unsigned   type      : 5;
-   unsigned   DPL       : 2;
-   unsigned   present   : 1;
-   unsigned   offset_hi : 16;
-} Gate;
-
-#define GATE_OFFSET(_gate)       (((_gate).offset_hi << 16) | (_gate).offset_lo)
-
-#define GATE_OFFSET_LO(_dw)      (((int)(_dw)) & 0xffff)
-#define GATE_OFFSET_HI(_dw)      ((((int)(_dw)) >> 16) & 0xffff)
-
-#define TASK_GATE           0x05
-#define INTER_GATE          0x0e
-#define TRAP_GATE           0x0f
-
-
-/*
- *   descriptor table registers
- */
-
-/* Need to pack the DTR struct so the
-   offset starts right after the limit */
-
-#ifdef __GNUC__
-
-typedef struct DTR {
-   uint16 limit;
-   uint32 offset __attribute__ ((packed));
-} DTR;
-
-#elif _MSC_VER
-
-#pragma pack(push)
-#pragma pack(2)
-
-typedef struct {
-   uint16 limit;
-   uint32 offset;
-} DTR;
-
-#pragma pack(pop)
-
-#else
-# error
-#endif
-
-typedef union {
-   DTR dtr;
-   uint32 word[2];
-} DTRWords;
-
 typedef struct {
    uint32 va;
    Selector seg;
 } FarPtr;
 
-
-/*
- *   far pointers
- */
-
 typedef struct FarPtr16 {
    uint16   offset;
    uint16   selector;
@@ -489,38 +414,66 @@ typedef struct FarPtr32 {
 } FarPtr32;
 
 
-
-
 /*
- *   register numbering
+ * General-purpose registers and segment registers 
  */
 
-#define REG_EAX         0x00
-#define REG_ECX         0x01
-#define REG_EDX         0x02
-#define REG_EBX         0x03
-#define REG_ESP         0x04
-#define REG_EBP         0x05
-#define REG_ESI         0x06
-#define REG_EDI         0x07
+typedef enum RegisterName {
+   REG_NULL = -1, // -1 so that NUM_REGS doesn't include REG_NULL
+   REG_EAX,
+   REG_ECX,
+   REG_EDX,
+   REG_EBX,
+   REG_ESP,
+   REG_EBP,
+   REG_ESI,
+   REG_EDI,
+   NUM_REGS
+} RegisterName;
 
 
-#define SEG_ES       0
-#define SEG_CS       1
-#define SEG_SS       2
-#define SEG_DS       3
-#define SEG_FS       4
-#define SEG_GS       5
-#define SEG_LDTR     6
-#define SEG_TR       7
+/* Return TRUE iff "reg" is a valid register number. */
+static INLINE Bool 
+x86_IsGoodRegister(RegisterName regName) 
+{ 
+   return (REG_NULL < regName && regName < NUM_REGS); 
+}
 
-#define NUM_SEGS     8
-   
+typedef enum SegmentName {
+   SEG_NULL = -1, // -1 so that NUM_SEGS doesn't include SEG_NULL
+   SEG_ES,
+   SEG_CS,
+   SEG_SS,
+   SEG_DS,
+   SEG_FS,
+   SEG_GS,
+   SEG_LDTR,
+   SEG_TR,
+   NUM_SEGS
+} SegmentName;
+
+#define NUM_NORMAL_SEGS  6
+
+static INLINE Bool 
+x86_IsGoodSegmentName(SegmentName segName)
+{ 
+   return SEG_NULL < segName && segName < NUM_SEGS; 
+}
+
+/* Return FALSE iff "segName" refers to the special SEG_LDTR or SEG_TR. */
+static INLINE Bool 
+x86_IsNormalSegmentName(SegmentName segName)
+{
+#ifdef ASSERT
+   ASSERT(x86_IsGoodSegmentName(segName));
+#endif
+   return segName < NUM_NORMAL_SEGS;
+}
+
+  
 /*
- *   exception frame.
+ * Exception frame.
  */
-
-
 typedef struct ExcFrame {
    /* pushed by sw */
    uint16      es, __esu;
@@ -573,7 +526,7 @@ typedef struct ExcFrame {
 #define PTE_PS          0x00000080
 #define PTE_G           0x00000100
 
-#define PTE_FLAGS       (PTE_P|PTE_RW|PTE_US)
+#define PTE_FLAGS       (PTE_P|PTE_RW|PTE_US|PTE_G)
 
 #define PTE_PAGE_TABLE  (PTE_P|PTE_RW|PTE_US|PTE_A|PTE_D) 
 
@@ -581,117 +534,65 @@ typedef struct ExcFrame {
 #define PTE_WRITEABLE(_pte) (((_pte) & PTE_RW) != 0)
 #define PTE_DIRTY(_pte)     (((_pte) & PTE_D) != 0)
 #define PTE_USER(_pte)      (((_pte) & PTE_US) != 0)
+#define PTE_GLOBAL(_pte)    (((_pte) & PTE_G) != 0)
 
-#define PTE_AVAIL_MASK  0xe00
-#define PTE_AVAIL_SHIFT 9
-#define PTE_PFN_MASK    0xfffff000
-#define PTE_PFN_SHIFT   12
-#define PDE_PFN_MASK    0xffc00000
+#define PTE_AVAIL_MASK       0xe00
+#define PTE_AVAIL_SHIFT      9
+#define PTE_PFN_MASK         0xfffff000
+#define PAE_PTE_PFN_MASK     0xffffff000LL
+#define PTE_PFN_SHIFT        12
+#define PAE_PTE_PFN_SHIFT    12
+#define PTE_2_PFN(_pte)      (((_pte) & PTE_PFN_MASK) >> PTE_PFN_SHIFT)
+#define PAE_PTE_2_PFN(_pte)  (((_pte) & PAE_PTE_PFN_MASK) >> PAE_PTE_PFN_SHIFT)
 
-#define PTE_2_PFN(_pte) (((_pte) & PTE_PFN_MASK) >> PTE_PFN_SHIFT)
-#define PDE_2_PFN(_pte) (((_pte) & PDE_PFN_MASK) >> PTE_PFN_SHIFT)
+#define PDE_PFN_MASK         0xffc00000
+#define PAE_PDE_PFN_MASK     0xfffe00000LL
+#define PDE_PFN_SHIFT        PTE_PFN_SHIFT
+#define PAE_PDE_PFN_SHIFT    PAE_PTE_PFN_SHIFT
 
+/* Use these only for 4M or 2M pde's, not the 4K ones. */
+#define PDE_2_PFN(_pte)      (((_pte) & PDE_PFN_MASK) >> PTE_PFN_SHIFT)
+#define PAE_PDE_2_PFN(_pde)  (((_pde) & PAE_PDE_PFN_MASK) >> PAE_PDE_PFN_SHIFT)
+
+#define PDPTE_PFN_MASK       0xffffff000LL
+#define PDPTE_PFN_SHIFT      12
+#define PDPTE_2_PFN(_p)      (((_p) & PDPTE_PFN_MASK) >> PDPTE_PFN_SHIFT)
 
 #define PDES_PER_PMAP      (PAGE_SIZE/sizeof(PDE))
 #define PTES_PER_PDIR      (PAGE_SIZE/sizeof(PTE))
 
-#define PFN_2_PDOFF(_a)   ((_a) >> 10)
-#define PFN_2_PGOFF(_a)   ((_a) & 0x3ff)
+#define PAE_PGOFF_MASK        0x1ff
+#define PAE_PDOFF_MASK        0x1ff
+#define PAE_RANGE_MASK        0x7ff
+
+#define PFN_2_PDOFF(_a)       ((_a) >> 10)
+#define PAE_PFN_2_PDOFF(_a)   (((_a) >> 9) & PAE_PDOFF_MASK)
+#define PFN_2_PGOFF(_a)       ((_a) & 0x3ff)
+#define PAE_PFN_2_PGOFF(_a)   ((_a) & PAE_PGOFF_MASK)
 
 #define PA_2_PDOFF(_a)   ((_a) >> (10+PAGE_SHIFT))
 #define PA_2_PGOFF(_a)   ((PA_2_PFN(_a)) & 0x3ff)
 
+#define PDOFF_2_LA(_p)            ((_p) << 22)
+#define LA_2_PDOFF(_a)            ((_a) >> 22)
+#define PAE_LA_2_PDOFF(_a)        (((_a) >> 21) & PAE_PDOFF_MASK)
+#define LA_2_PGOFF(_a)            (PFN_2_PGOFF(VA_2_VPN(_a)))
+#define LA_2_RANGE(_a)            LA_2_PDOFF((_a))
+#define PAE_LA_2_RANGE(_a)        (((_a) >> 21) & PAE_RANGE_MASK)
+#define PAE_LA_2_PDPTOFF(_a)      (((_a) >> 30) & 0x3)
 
+/* This following macro only work for flat model were VA==LA */
 #define VA_2_PDOFF(_a)    ((_a) >> 22)
 #define VA_2_PGOFF(_a)    (PFN_2_PGOFF(VA_2_VPN(_a)))
 
 #define PTE_2_PA(_a)      (PPN_2_PA(PTE_2_PFN(_a)))
 #define PDE_2_PA(_a)      (PPN_2_PA(PDE_2_PFN(_a)))
 
-
-/*
- * Location of the APICs' registers.
- */
-
-#define APIC_DEFAULT_ADDRESS      0xfee00000
-#define IOAPIC_DEFAULT_ADDRESS    0xfec00000
-
-#define IOAPIC_NUM_REDIR_REGS     24
-
-#define APICR_ID         0x02
-#define APICR_VERSION    0x03
-#define APICR_TPR        0x08
-#define APICR_APR        0x09
-#define APICR_PPR        0x0a
-#define APICR_EOI        0x0b
-#define APICR_LDR        0x0d
-#define APICR_DFR        0x0e
-#define APICR_SVR        0x0f
-#define APICR_ISR        0x10
-#define APICR_TMR        0x18
-#define APICR_IRR        0x20
-#define APICR_ESR        0x28
-#define APICR_ICRLO      0x30
-#define APICR_ICRHI      0x31
-#define APICR_TIMERLVT   0x32
-#define APICR_PCLVT      0x34
-#define APICR_LVT0       0x35
-#define APICR_LVT1       0x36
-#define APICR_ERRLVT     0x37
-#define APICR_INITCNT    0x38
-#define APICR_CURCNT     0x39
-#define APICR_DIVIDER    0x3e
-
-#define APICR_SIZE       0x400
-
-#define APIC_PR_MASK          0x000000ff
-#define APIC_PR_XMASK         0x000000f0
-#define APIC_PR_YMASK         0x0000000f
-#define APIC_PR_X(_r)         (((_r) & APIC_PR_XMASK) >> 4)
-#define APIC_PR_Y(_r)         (((_r) & APIC_PR_YMASK) >> 0)
-
-#define APIC_SVR_ONES         0x0000000f
-#define APIC_SVR_VECTOR       0x000000f0
-#define APIC_SVR_APICENABLE   0x00000100
-#define APIC_SVR_FOCUSCHECK   0x00000200
-
-#define APIC_BASE_MSR         0x1b
-
-#define APIC_MSR_BASEMASK     0xfffff000
-#define APIC_MSR_ENABLED      0x00000800
-#define APIC_MSR_BSP          0x00000100
-
-#define APIC_VTE_VECTOR_MASK  0x000000ff
-#define APIC_VTE_MODE_FIXED   0x00000000
-#define APIC_VTE_MODE_NMI     0x00000400
-#define APIC_VTE_MODE_EXTINT  0x00000700
-#define APIC_VTE_MODE_MASK    0x00000700
-#define APIC_VTE_DELVSTATUS   0x00001000
-#define APIC_VTE_PINPOL       0x00002000
-#define APIC_VTE_REMIRR       0x00004000
-#define APIC_VTE_TRIGMODE     0x00008000
-#define APIC_VTE_MASK         0x00010000
-#define APIC_VTE_TIMERMODE    0x00020000
-
-/* Kernels 2.4.0-test8 and higher have the same definition */
-#ifndef APIC_ID_MASK
-#   define APIC_ID_MASK          0x0f000000
-#endif
-#define APIC_ID_SHIFT         24
-
-#define IOAPICID              0x00
-#define IOAPICVER             0x01
-#define IOAPICARB             0x02
-#define IOREDTBL_FIRST        0x10
-#define IOREDTBL_LAST         0x3f
-
 /*
  *   exception error codes
  */
 
 #define EXC_NONE         -1
-#define EXC_BUS_ERROR    -2   /* XXX ? */
-
 #define EXC_DE            0
 #define EXC_DB            1
 #define EXC_NMI           2
@@ -769,15 +670,15 @@ typedef struct DebugControlRegister {
   
 
                  
-#define MAKE_SELECTOR(_index, _ti, _RPL)   \
+#define MAKE_SELECTOR(_index, _ti, _RPL)                     \
 ( (((_index) << SELECTOR_INDEX_SHIFT) & SELECTOR_INDEX_MASK) \
-  | (((_ti)    << SELECTOR_TI_SHIFT)    & SELECTOR_TI_MASK) \
-  | (((_RPL)   << SELECTOR_RPL_SHIFT)   & SELECTOR_RPL_MASK) )
+  | (((_ti)  << SELECTOR_TI_SHIFT)    & SELECTOR_TI_MASK)    \
+  | (((_RPL) << SELECTOR_RPL_SHIFT)   & SELECTOR_RPL_MASK) )
 
 
 #define MAKE_PTE(_mpfn, _avail, _flags) \
                         (((_mpfn) << PTE_PFN_SHIFT) \
-                         | (((_avail) & PTE_AVAIL_MASK) << PTE_AVAIL_SHIFT) \
+                         | (((_avail) << PTE_AVAIL_SHIFT) & PTE_AVAIL_MASK) \
                          | (_flags))
 
 
@@ -794,9 +695,9 @@ typedef struct DebugControlRegister {
  * Intel hardware.
  */
 typedef struct SMMSegState {
-  uint32 limit;   /* Limit from descriptor */
-  uint32 base;    /* Base from descriptor */
-  uint32 rights;  /* Rest of the descriptor */
+   uint32 limit;   /* Limit from descriptor */
+   uint32 base;    /* Base from descriptor */
+   uint32 rights;  /* Rest of the descriptor */
 } SMMSegState;
 
 /*
@@ -805,28 +706,32 @@ typedef struct SMMSegState {
  */
 
 typedef struct SMMStateSaveMap {
-  uint8  reserved0[248];  
-  uint32 smbase;
-  uint32 revId; 
-  uint16 ioInstRestart;
-  uint16 haltRestart;
-  uint8  reserved01[12];
-  uint32 ioInstEIP; 
-  uint8  reserved1[18];
-  uint32 cr4reg;
-  uint8  reserved2[4];
-  SMMSegState hiddenSegPart[7];
-  SMMSegState gdtr;
-  SMMSegState idtr;
-  SMMSegState hiddenTRpart;
-  uint32 segRegs[8];
-  Reg    dr7reg;
-  Reg    dr6reg;
-  Reg    gpRegs[8];
-  Reg    eip;
-  Reg    eflags;
-  Reg    cr3reg;
-  Reg    cr0reg;
+   union {
+      uint32 selState;       // We steal a enough of this (hopefully) unused
+                             // space to stash away the selector state. 
+      uint8  reserved0[248];
+   } u;
+   uint32 smbase;
+   uint32 revId; 
+   uint16 ioInstRestart;
+   uint16 haltRestart;
+   uint8  reserved01[12];
+   uint32 ioInstEIP; 
+   uint8  reserved1[18];
+   uint32 cr4reg;
+   uint8  reserved2[4];
+   SMMSegState hiddenSegPart[7];
+   SMMSegState gdtr;
+   SMMSegState idtr;
+   SMMSegState hiddenTRpart;
+   uint32 segRegs[8];
+   Reg    dr7reg;
+   Reg    dr6reg;
+   Reg    gpRegs[8];
+   Reg    eip;
+   Reg    eflags;
+   Reg    cr3reg;
+   Reg    cr0reg;
 } SMMStateSaveMap;
 
 #define SMM_SAVEMAP_SIZE   512
@@ -836,5 +741,45 @@ typedef struct SMMStateSaveMap {
 
 #define SMM_RIGHTS_BITS   0x00f0ff00
 
+/*
+ * FPU FSAVE/FXSAVE area format. 
+ *
+ * fsave_hard_struct - Memory format of the FSAVE/FNSAVE instruction. 108 bytes
+ * fxsave_hard_struct - Memory format of the FXSAVE instruction. 512 bytes
+ *                    Note: fxsave_hard_struct must be 128bit aligned. 
+ *
+ * Taken from linux headers
+ */
+
+struct fsave_hard_struct {
+   long   cwd;
+   long   swd;
+   long   twd;
+   long   fip;
+   long   fcs;
+   long   foo;
+   long   fos;
+   long   st_space[20];   /* 8*10 bytes for each FP-reg = 80 bytes */
+};
+
+struct fxsave_hard_struct {
+   unsigned short fxcwd;
+   unsigned short fxswd;
+   unsigned short fxtwd;
+   unsigned short fxfopcode;
+   long     fxfip;
+   short    fxfcs;
+   short    __reserved_00;
+   long     fxfoo;
+   short    fxfos;
+   short    __reserved_01;
+   long     mxcsr;
+   long     __reserved_02;
+   long     st_space[32];     /* 8*16 bytes for each FP/MMX-reg = 128 bytes */
+   long     xmm_space[32];    /* 8*16 bytes for each XMM-reg = 128 bytes */
+   long     __reserved_03 [14*4]; /* 14 16byte lines for remainder */
+};
+
+
 #endif /* _X86_H_ */
 
