@@ -11,6 +11,13 @@
 
 struct VMDriver;
 
+typedef struct MemRange {
+	struct MemRange *next;
+	VA kAddr;
+	PA pAddr;
+	unsigned long numPages;
+} MemRange;
+
 struct vmx86_softc {
 	struct VMDriver *vm_vm;
 	struct selinfo vm_rsel;
@@ -18,6 +25,8 @@ struct vmx86_softc {
 	int vm_flags;
 	dev_t vm_dev;
 	struct vmmon_softc *vm_monsc;
+	VmTimeStart startTime;
+	MemRange vm_low_pages;
 };
 
 #define VMFL_SELTIMO	0x01
