@@ -80,6 +80,7 @@
 #include "vmx86.h"
 #include "initblock.h"
 #include "task.h"
+#include "cpuid.h"
 #include "hostif.h"
 #include "driver.h"
 #include "speaker_reg.h"
@@ -108,6 +109,11 @@ static struct vmx86_softc * vm_allocate(struct vmmon_softc *);
 static void vm_destroy(struct vmmon_softc *, int);
 static void vm_deallocate(struct vmx86_softc *);
 
+static int vmmon_alloc_low_pages(struct vmx86_softc *, unsigned long, PA *);
+static MemRange * vmmon_remove_low_pages(MemRange *, PA);
+static int vmmon_free_low_pages(struct vmx86_softc *, PA);
+static void vmmon_destroy_low_pages(MemRange *);
+
 struct vmmon_softc vmmon_sc;
 
 static struct cdevsw vmmon_dev = {
@@ -143,6 +149,8 @@ vmmon_handle(struct lkm_table *lkmtp, int cmd)
 	case LKM_E_LOAD:
 		if (lkmexists(lkmtp)) 
 			return EEXIST;
+		CPUID_Init();
+		Task_ProbeSysenterMSR();
 		break;
 		
 	case LKM_E_UNLOAD:
@@ -168,7 +176,6 @@ vmmon_open(dev_t dev, int flag, int mode, struct proc *p)
 	struct vmmon_softc *vmmonsc;
 	struct vmx86_softc *vmxsc;
 	int error;
-	VMDriver *vm;
 
 	if (p->p_dupfd >= 0)
 		return ENODEV;
@@ -197,14 +204,12 @@ vmmon_open(dev_t dev, int flag, int mode, struct proc *p)
 		return error;
 	}
 
-	vm = Vmx86_Init((void *)vmxsc, (void *)(p->p_pid));
-	if (vm == NULL) {
-		vm_destroy(vmmonsc, VMNUM(vmxsc->vm_dev));
-		error = ENOMEM;
-	} else {
-		vmxsc->vm_vm = vm;
-		callout_init(&vmxsc->vm_callout);
-	}
+	/*
+	 * Snap shot the time stamp counter and the real time so we
+	 * can later compute an estimate of the cycle time.
+	 */
+	Vmx86_SetStartTime(&vmxsc->startTime);
+	callout_init(&vmxsc->vm_callout);
 
 	return error;
 }
@@ -216,6 +221,7 @@ vmmon_close(dev_t dev, int flags, int mode, struct proc *p)
 	int num;
 	struct vmmon_softc *sc;
 	struct vmx86_softc *vmxsc;
+	MemRange *head, *first;
 
 	VMDEBUG("vmmon: close vm %d by pid %d\n", VMNUM(dev), p->p_pid);
 
@@ -229,8 +235,18 @@ vmmon_close(dev_t dev, int flags, int mode, struct proc *p)
 
 	callout_stop(&vmxsc->vm_callout);
 
+	/*
+	 * Destroy all the remaining ranges.
+	 */
+	for (head = &vmxsc->vm_low_pages; (first = head->next) != head; ) {
+		/* Unlink the first item at the beginning of the list. */
+		head->next = first->next;
+
+		vmmon_destroy_low_pages(first);
+	}
+
 	if (vmxsc->vm_vm != NULL)
-		Vmx86_DestroyVM(vmxsc->vm_vm);
+		Vmx86_ReleaseVM(vmxsc->vm_vm);
 
 	vm_destroy(sc, num);
 
@@ -286,6 +302,175 @@ vmmon_fake_clonedev(dev_t dev, int flag, struct proc *p)
 	return ENXIO;
 }
 
+/*
+ * Allocate and lock numPages of memory physically continguous addresses
+ * below 4GB.
+ * The requirement should be < 4GB but there is code in the vmware binaries
+ * which relys on < 16MB to fix a bug. c.f. ../linux/driver.c
+ */
+static int
+vmmon_alloc_low_pages(struct vmx86_softc *vmxsc, unsigned long numPages, PA *addr)
+{
+	MemRange *r;
+	struct pglist l;
+	struct vm_page *m;
+	paddr_t high;
+	paddr_t curaddr, lastaddr;
+	vaddr_t va;
+	int error, n;
+	extern paddr_t avail_end;
+#if 0
+#define ISA_DMA_BOUNCE_THRESHOLD (16UL * 1024 * 1024)
+#else
+	/*
+	 * Unfortunately, all the low memory seems to be taken in multiuser
+	 * mode and uvm_pglistalloc doesn't try to free some for us.
+	 * So just use any pages we can get.
+	 */
+#define ISA_DMA_BOUNCE_THRESHOLD (3072UL * 1024 * 1024)
+#endif
+
+	if (numPages == 0)
+		return EINVAL;
+
+	r = malloc(sizeof *r, M_DEVBUF, M_WAITOK);
+	if (r == NULL) {
+		VMDEBUG("vmmon_alloc_low_pages: malloc MemRange failed\n");
+		return ENOMEM;
+	}
+	if (avail_end > ISA_DMA_BOUNCE_THRESHOLD)
+		high = trunc_page(ISA_DMA_BOUNCE_THRESHOLD);
+	else
+		high = trunc_page(avail_end);
+	/*
+	 * Allocate pages from the VM system.
+	 */
+	TAILQ_INIT(&l);
+	VMDEBUG("vmmon_alloc_low_pages: pglistalloc: sz %lu hi %lu\n",
+	    numPages * PAGE_SIZE, high);
+	error = uvm_pglistalloc(numPages * PAGE_SIZE, 0, high, PAGE_SIZE, 0,
+		    &l, 1, 1);
+	if (error != 0) {
+		VMDEBUG("vmmon_alloc_low_pages: pglistalloc failed: %d\n", error);
+		return error;
+	}
+	/*
+	 * Check that we got a single physically continguous set of pages.
+	 */
+	m = l.tqh_first;
+	lastaddr = r->pAddr = VM_PAGE_TO_PHYS(m);
+	r->numPages = 1;
+	m = m->pageq.tqe_next;
+
+	for (; m != NULL; m = m->pageq.tqe_next) {
+		curaddr = VM_PAGE_TO_PHYS(m);
+#ifdef DIAGNOSTIC
+		if (curaddr >= high) {
+			VMDEBUG("vmmon_alloc_low_pages: uvm_pglistalloc returned non-sensical"
+			    " address 0x%lx\n", curaddr);
+			goto bad;
+		}
+#endif
+		if (curaddr == (lastaddr + PAGE_SIZE))
+			r->numPages++;
+		else {
+			VMDEBUG("vmmon_alloc_low_pages: non-continguous pages\n");
+			goto bad;
+		}
+		lastaddr = curaddr;
+	}
+
+	/*
+	 * Map the segment into wired kernel virtual memory.
+	 */
+	va = uvm_km_valloc(kernel_map, numPages * PAGE_SIZE);
+	if (va == 0) {
+		VMDEBUG("vmmon_alloc_low_pages: km_valloc failed\n");
+		goto bad;
+	}
+	r->kAddr = va;
+
+	for (curaddr = r->pAddr, n = r->numPages;
+	     n != 0;
+	     curaddr += PAGE_SIZE, va += PAGE_SIZE, n--) {
+		pmap_enter(pmap_kernel(), va, curaddr,
+		    VM_PROT_READ | VM_PROT_WRITE,
+		    PMAP_WIRED | VM_PROT_READ | VM_PROT_WRITE);
+	}
+	pmap_update(pmap_kernel());
+
+	*addr = r->pAddr;
+
+	/* XXX should lock list insertion. */
+	r->next = vmxsc->vm_low_pages.next;
+	vmxsc->vm_low_pages.next = r;
+
+	return 0;
+
+bad:
+	uvm_pglistfree(&l);
+	free(r, M_DEVBUF);
+	return ENOMEM;
+}
+
+static MemRange *
+vmmon_remove_low_pages(MemRange *head, PA addr)
+{
+	MemRange *cur, *next;
+
+	for (cur = head; (next = cur->next) != head; cur = next) {
+		if (next->pAddr == addr) {
+			/* Unlink the matching item. */
+			cur->next = next->next;
+			return next;
+		}
+	}
+
+	return NULL;
+}
+static int
+vmmon_free_low_pages(struct vmx86_softc *vmxsc, PA addr)
+{
+	MemRange *r;
+
+	/* XXX should protect this call by lock */
+	r = vmmon_remove_low_pages(&vmxsc->vm_low_pages, addr);
+	if (r == NULL)
+		return EINVAL;
+
+	vmmon_destroy_low_pages(r);
+	return 0;
+}
+
+static void
+vmmon_destroy_low_pages(MemRange *r)
+{
+	struct vm_page *m;
+	struct pglist l;
+	paddr_t addr;
+	int n;
+
+	/*
+	 * Unmap the kernel virtual memory used by this range.
+	 */
+	uvm_km_free(kernel_map, (vaddr_t)r->kAddr, r->numPages * PAGE_SIZE);
+
+	/*
+	 * Build a list of pages to free back to the VM system.
+	 */
+	TAILQ_INIT(&l);
+	for (addr = r->pAddr, n = r->numPages;
+	    n != 0;
+	    addr += PAGE_SIZE, n--) {
+		m = PHYS_TO_VM_PAGE(addr);
+		TAILQ_INSERT_TAIL(&l, m, pageq);
+	}
+
+	uvm_pglistfree(&l);
+
+	free(r, M_DEVBUF);
+}
+
 /*
  * This function uses the hack to make PTIOCLINUX ioctls (passthroughs
  * from the emulation) return EJUSTRETURN to be able to have them
@@ -297,14 +482,20 @@ vmmon_ioctl(dev_t dev, u_long cmd, caddr_t data, int flag, struct proc *p)
 	struct vmx86_softc *vmxsc;
 	struct vmmon_softc *sc;
 	int error, num;
+	unsigned long numPages;
 	struct ioctl_pt *pt;
-	InitBlock params;
+	WslimitsInfo wslimitsInfo;
+	InitBlock initParams;
+	MA ma;
+	PA pa;
 	MPN mpn;
 	int32_t limit;
-	VMGetSetMemInfoInArgs memargs;
-	VMGetVMInfoResult rinfo;
-	VMSetVMInfoArgs sinfo;
+	VMSetMemInfoArgs memargs;
 	VMDriver *vm;
+	Bool setVMPtr, probe;
+#ifdef USE_PERFCOUNTERS
+	Bool forceEnable;
+#endif
 
 	VMDEBUG("vmmon: ioctl %lx on vm %d by pid %d\n",
 	    cmd, VMNUM(dev), p->p_pid);
@@ -322,79 +513,118 @@ vmmon_ioctl(dev_t dev, u_long cmd, caddr_t data, int flag, struct proc *p)
 	pt = (struct ioctl_pt *)data;
 
 	vm = vmxsc->vm_vm;
-	/* XXX strange error to return, but it's what the Linux module does */
-	if (vm == NULL && pt->com != IOCTLCMD_GET_MEM_INFO)
-		return EDEADLK;
 
 	VMDEBUG("vmmon: ioctl cmd %lu\n", pt->com);
 
 	switch (pt->com) {
-	case IOCTLCMD_INIT:
-		error = copyin(pt->data, &params, sizeof params);
+	case IOCTLCMD_VERSION:
+		pt->data = (void *)VMMON_VERSION;
+		return EJUSTRETURN;
+	case IOCTLCMD_CREATE_VM:
+		if (vm != NULL)
+			return EINVAL;
+		error = copyin(pt->data, &wslimitsInfo, sizeof wslimitsInfo);
 		if (error != 0)
 			return error;
-		if (Vmx86_CreateVM(vm, &params) != 0)
-			return EIO;
-		error = copyout(&params, pt->data, sizeof params);
+		vm = Vmx86_CreateVM((void *)&vmxsc, (void*)p->p_pid,
+		    &wslimitsInfo);
+		if (vm == NULL)
+			return ENOMEM;
+		vmxsc->vm_vm = vm;
+		pt->data = (void *) vm->id;
+		return EJUSTRETURN;
+	case IOCTLCMD_BIND_VM:
+		if (vm != NULL)
+			return EINVAL;
+		vm = Vmx86_BindVM((int)pt->data);
+		if (vm == NULL)
+			return EINVAL;
+		vmxsc->vm_vm = vm;
+		break;
+	case IOCTLCMD_RELEASE_VM:
+		if (vm == NULL)
+			return EINVAL;
+		vmxsc->vm_vm = NULL;
+		Vmx86_ReleaseVM(vm);
+		break;
+	case IOCTLCMD_INIT_VM:
+		if (vm == NULL)
+			return EINVAL;
+		error = copyin(pt->data, &initParams, sizeof initParams);
+		if (error != 0)
+			return error;
+		if (Vmx86_InitVM(vm, &initParams) != 0)
+			return EINVAL;
+		error = copyout(&initParams, pt->data, sizeof initParams);
 		if (error != 0)
 			return error;
 		break;
-	case IOCTLCMD_LATE_INIT:
-		Vmx86_LateInit(vm);
+	case IOCTLCMD_LATE_INIT_VM:
+		if (vm == NULL)
+			return EINVAL;
+		if (Vmx86_LateInitVM(vm) != 0 )
+			return EINVAL;
 		break;
-	case IOCTLCMD_RUN:
+	case IOCTLCMD_RUN_VM:
+		if (vm == NULL)
+			return EINVAL;
 		pt->data = (void *)Vmx86_RunVM(vm);
 		return EJUSTRETURN;
-	case IOCTLCMD_LOOKUPMPN:
+	case IOCTLCMD_LOOK_UP_MPN:
 		mpn = HostIF_LookupUserMPN(vm, pt->data);
 		pt->data = (void *)mpn;
 		return EJUSTRETURN;
-	case IOCTLCMD_BEEP:
-		sysbeep(1500, hz/5);
-		break;
-	case IOCTLCMD_LOCKPAGE:
+	case IOCTLCMD_LOCK_PAGE:
+		if (vm == NULL)
+			return EINVAL;
 		pt->data = (void *)Vmx86_LockPage(vm, pt->data, TRUE);
 		return EJUSTRETURN;
-	case IOCTLCMD_UNLOCKPAGE:
+	case IOCTLCMD_UNLOCK_PAGE:
+		if (vm == NULL)
+			return EINVAL;
 		pt->data = (void *)Vmx86_UnlockPage(vm, pt->data, TRUE);
 		return EJUSTRETURN;
-	case IOCTLCMD_ISMPSAFE:
-#ifdef MULTIPROCESSOR
-		pt->data = (void *)TRUE;
-#else
-		pt->data = (void *)FALSE;
-#endif
-		return EJUSTRETURN;
-	case IOCTLCMD_APICBASE:
-		/*
-		 * Hardcode probe to false vor vmware-2.0.
-		 */
-		pt->data = (void *)HostIF_APIC_Base(vm,
-					    (Bool)(u_long)pt->data, FALSE);
-		return EJUSTRETURN;
-	case IOCTLCMD_IOAPICBASE:
-		pt->data = (void *)HostIF_IOAPIC_Base(vm);
-		return EJUSTRETURN;
 	case IOCTLCMD_GET_NUM_VMS:
 		pt->data = (void *)Vmx86_GetNumVMs();
 		return EJUSTRETURN;
-	case IOCTLCMD_DECLARE_SLAVE:
-		break;
-	case IOCTLCMD_CHECK_MEMORY:
-		pt->data = (void *)HostIF_CheckMemory(vm);
+	case IOCTLCMD_APIC_BASE:
+		if (vm == NULL)
+			return EINVAL;
+		setVMPtr = ((int)pt->data & APIC_FLAG_DISABLE_NMI) != 0;
+		probe = ((int)pt->data & APIC_FLAG_PROBE) != 0;
+#ifdef USE_PERFCOUNTERS
+		forceEnable = ((int)pt->data & APIC_FLAG_FORCE_ENABLE) != 0;
+#endif
+		ma = HostIF_APICBase(vm, setVMPtr, probe);
+#ifdef USE_PERFCOUNTERS
+		if ((ma == 0) && forceEnable) {
+			ma = HostIF_APICEnable(vm);
+		}
+#endif
+		pt->data = (void *)ma;
 		return EJUSTRETURN;
-	case IOCTLCMD_ALLOW_CORE_DUMP:
-		break;
-	case IOCTLCMD_GETSTATS:
+	case IOCTLCMD_IOAPIC_BASE:
+		if (vm == NULL)
+			return EINVAL;
+		pt->data = (void *)HostIF_IOAPICBase(vm);
+		return EJUSTRETURN;
+	case IOCTLCMD_GET_STATS:
+		if (vm == NULL)
+			return EINVAL;
 		error = copyout(&vm->stats, pt->data, sizeof vm->stats);
 		if (error != 0)
 			return error;
 		break;
-	case IOCTLCMD_SETSTATS:
+	case IOCTLCMD_SET_STATS:
+		if (vm == NULL)
+			return EINVAL;
 		error = copyin(pt->data, &vm->stats, sizeof vm->stats);
 		if (error != 0)
 			return error;
 		break;
+	case IOCTLCMD_GET_HARD_LIMIT:
+		pt->data = (void *)Vmx86_GetLockedPagesLimit();
+		return EJUSTRETURN;
 	case IOCTLCMD_SET_HARD_LIMIT:
 		error = copyin(pt->data, &limit, sizeof limit);
 		if (error != 0)
@@ -402,46 +632,52 @@ vmmon_ioctl(dev_t dev, u_long cmd, caddr_t data, int flag, struct proc *p)
 		if (!Vmx86_SetLockedPagesLimit(limit))
 			return EINVAL;
 		break;
-	case IOCTLCMD_GET_HARD_LIMIT:
-		pt->data = (void *)Vmx86_GetLockedPagesLimit();
-		return EJUSTRETURN;
 	case IOCTLCMD_GET_MEM_INFO:
+		if (vm == NULL)
+			return EINVAL;
 		if (!Vmx86_GetMemInfoCopy(vm,
-		    (VMGetSetMemInfoOutArgs *)pt->data))
+		    (VMGetMemInfoArgs *)pt->data))
 			return EINVAL;
 		break;
 	case IOCTLCMD_SET_MEM_INFO:
+		if (vm == NULL)
+			return EINVAL;
 		error = copyin(pt->data, &memargs, sizeof memargs);
 		if (error != 0)
 			return error;
 		Vmx86_SetMemInfo(vm, &memargs);
 		break;
-	case IOCTLCMD_GET_VM_LIST:
-		if (!Vmx86_GetVMListCopy((VMGetVMListResult *)pt->data))
-			return EINVAL;
-		break;
-	case IOCTLCMD_GET_VM_INFO:
-		error = copyin(pt->data, &rinfo, sizeof rinfo);
-		if (error != 0)
-			return error;
-		if (!Vmx86_GetVMInfo(rinfo.vmUID, &rinfo))
-			return EINVAL;
-		error = copyout(&rinfo, pt->data, sizeof rinfo);
-		if (error != 0)
-			return error;
-		break;
-	case IOCTLCMD_SET_VM_INFO:
-		error = copyin(pt->data, &sinfo, sizeof sinfo);
-		if (error != 0)
-			return error;
-		Vmx86_SetVMInfo(vm, &sinfo);
-		break;
-	case IOCTLCMD_BROADCAST_IPI:
-		return ENOTTY;
-	case IOCTLCMD_FREE_RESOURCES:
-		pt->data = (void *)HostIF_FreeAllResources(vm);
+	case IOCTLCMD_PAE_ENABLED:
+		pt->data = (void *)(int)Vmx86_PAEEnabled();
 		return EJUSTRETURN;
-	case IOCTLCMD_SETUID:
+	case IOCTLCMD_GET_TOTAL_MEM_USAGE:
+		pt->data = (void *)Vmx86_GetTotalMemUsage();
+		return EJUSTRETURN;
+	case IOCTLCMD_SET_UID:
+		break;
+	case IOCTLCMD_IS_MP_SAFE:
+#ifdef MULTIPROCESSOR
+		pt->data = (void *)TRUE;
+#else
+		pt->data = (void *)FALSE;
+#endif
+		return EJUSTRETURN;
+	case IOCTLCMD_GET_MHZ_ESTIMATE:
+		pt->data = (void *)Vmx86_GetMHzEstimate(&vmxsc->startTime);
+		return EJUSTRETURN;
+	case IOCTLCMD_ALLOW_CORE_DUMP:
+		break;
+
+	case IOCTLCMD_CHECK_MEMORY:
+		if (vm == NULL)
+			return EINVAL;
+		pt->data = (void *)HostIF_CheckMemory(vm);
+		return EJUSTRETURN;
+
+	case IOCTLCMD_BROADCAST_IPI:	// SMP 2.2.8+ only
+		return ENOTTY;
+
+#ifdef SUPPORT_PASSTHROUGH
 	case IOCTLCMD_REGISTER_PASSTHROUGH_IO:
 	case IOCTLCMD_REGISTER_PASSTHROUGH_IRQ:
 	case IOCTLCMD_FREE_PASSTHROUGH_IO:
@@ -449,8 +685,57 @@ vmmon_ioctl(dev_t dev, u_long cmd, caddr_t data, int flag, struct proc *p)
 	case IOCTLCMD_START_PASSTHROUGH:
 	case IOCTLCMD_STOP_PASSTHROUGH:
 	case IOCTLCMD_QUERY_PASSTHROUGH:
-	case IOCTLCMD_BLUE_SCREEN:
-	case IOCTLCMD_GET_SET_MEM_INFO:
+		/* not yet supported */
+		return ENOTTY;
+#endif /* SUPPORT_PASSTHROUGH */
+
+#ifdef USE_PERFCOUNTERS
+	case IOCTLCMD_REGISTER_PERFCTR:
+	case IOCTLCMD_START_PERFCTR:
+	case IOCTLCMD_STOP_PERFCTR:
+	case IOCTLCMD_RELEASE_PERFCTR:
+		/* not yet supported */
+		return ENOTTY;
+#endif /* USE_PERFCOUNTERS */
+
+	case IOCTLCMD_ALLOC_LOW_PAGES:
+		numPages = (unsigned long)fuword(pt->data);
+		VMDEBUG("vmmon: alloc_low_pages %lu\n", numPages);
+		if ((int)numPages == -1)
+			return EINVAL;
+		error = vmmon_alloc_low_pages(vmxsc, numPages, &pa);
+		if (error != 0) {
+			VMDEBUG("vmmon: cannot alloc_low_pages: %d\n", error);
+			return error;
+		}
+		error = suword(pt->data, pa);
+		if (error != 0 ) {
+			VMDEBUG("vmmon: cannot suword: %d\n", error);
+			return error;
+		}
+		break;
+
+	case IOCTLCMD_FREE_LOW_PAGES:
+		VMDEBUG("vmmon: free_low_pages 0x%08lx\n", (unsigned long)pt->data);
+		error = vmmon_free_low_pages(vmxsc, (PA)pt->data);
+		if (error != 0)
+			return error;
+		return EJUSTRETURN;
+
+#ifdef VMX86_SERVER
+	case IOCTLCMD_ALLOC_MACHINE_MEMORY:
+	case IOCTLCMD_POWERON_SLACK_MEMORY:
+	case IOCTLCMD_ASYNC_ACTIONS:
+	case IOCTLCMD_FILE_OPEN:
+	case IOCTLCMD_FILE_CLOSE:
+	case IOCTLCMD_SET_MMAP_LAST:
+#endif
+#ifdef VMX86_SERVER
+	/* bora-root/fileutils/ls needs to be recompiled if the following
+	case definition ever changes */
+	case IOCTLCMD_VMFS_GET_LENGTH:
+	case IOCTLCMD_VMFS_PHYS_MEM_IO:
+#endif
 	default:
 		return ENOTTY;
 	}
@@ -521,6 +806,7 @@ vm_create(struct vmmon_softc *sc, struct vmx86_softc **vmxpp)
 	if (i >= sc->sc_maxvm)
 		sc->sc_maxvm = i;
 	vmxsc->vm_dev = MAKEVMDEV(sc->sc_dev, i);
+	vmxsc->vm_low_pages.next = &vmxsc->vm_low_pages;
 
 	*vmxpp = vmxsc;
 
@@ -675,4 +961,5 @@ Panic(char *fmt, ...)
   
 	exit1(curproc, 0);
 	/* NOTREACHED */
+	for(;;); /* to suppress gcc warning */
 }
