@@ -84,6 +84,7 @@ extern vm_map_t kernel_map;
 
 #ifdef SMP_GLOBAL_VMLOCK
 static struct simplelock vmmon_lock;
+static proc *vmmon_lock_task;
 static int vmmon_lock_holder;
 #endif
 
@@ -91,6 +92,7 @@ static int host_lock_ppn(PPN);
 static int host_unlock_ppn(PPN);
 static void UnlockEntry(void *, MemTrackEntry *);
 static void * FindFunc(void *, MemTrackEntry *);
+static INLINE MPN HostifVa2Mpn(VA);
 
 
 static
@@ -118,13 +120,15 @@ int host_unlock_ppn(PPN ppn)
 /*
  *----------------------------------------------------------------------
  *
- * FindMPN --
+ * HostifVa2Mpn --
+ *
+ *    Walks through the hardware page tables of the current process to try to
+ *    find the the Machine Page Number (i.e. page number in the host physical
+ *    address space) associated to a virtual address.
  *
- *      walks through the hardware defined page tables
- *      to find a valid mpn.
- *      
  * Results:
- *      mpn on success, 0 on failure
+ *    The MPN if the page is found in the tables
+ *    0 if the page is not found in the tables
  *
  * Side effects:
  *      None.
@@ -132,10 +136,9 @@ int host_unlock_ppn(PPN ppn)
  *----------------------------------------------------------------------
  */    
 static INLINE MPN
-FindMPN(PPN ppn)
+HostifVa2Mpn(VA addr)
 {
-	vaddr_t vaddr = (vaddr_t)VPN_2_VA(ppn);
-	pt_entry_t *pteptr = (pt_entry_t *)vtopte(vaddr);
+	pt_entry_t *pteptr = (pt_entry_t *)vtopte((vaddr_t)addr);
 	PTE pte;
    
 	pte = *pteptr;
@@ -189,30 +192,11 @@ HostIF_InitSpinLock(void)
 int
 HostIF_Init(VMDriver *vm)
 {  
-	vm->memtracker = MemTrack_Init(sizeof(MemTrackEntry));
+	vm->memtracker = MemTrack_Init();
 	vm->physTracker = PhysTrack_Init();
 	return (vm->memtracker == NULL || vm->physTracker==NULL);
 }
 
-/*
- *----------------------------------------------------------------------
- *
- * HostIF_IsSystemSMP --
- *
- *      define if we are running on a SMP machine
- *
- * Results:
- *      
- *      TRUE if SMP
- *      
- *----------------------------------------------------------------------
- */
-Bool    
-HostIF_IsSystemSMP(void)
-{
-	return FALSE;
-}
-
 /*
  *----------------------------------------------------------------------
  *
@@ -234,7 +218,7 @@ MPN
 HostIF_LookupUserMPN(VMDriver *vm, char * addr)
 {
 	MPN mpn;
-	mpn = FindMPN(PTR_2_VPN(addr));
+	mpn = HostifVa2Mpn((VA)addr);
 	if (mpn == 0) {
 		NOT_TESTED();
 	}
@@ -300,10 +284,10 @@ HostIF_LockPage(VMDriver *vm, void *addr)
 	MPN mpn;
 	VPN vpn;
 	MemTrackEntry *entryPtr;
-	int c;
+	volatile int c;
 
 	vpn = PTR_2_VPN(addr);
-	entryPtr = (MemTrackEntry *) MemTrack_LookupVPN(vm->memtracker, vpn);
+	entryPtr = MemTrack_LookupVPN(vm->memtracker, vpn);
 
 	if ((entryPtr == NULL) || (entryPtr->mpn == 0)) {
 		/*
@@ -311,17 +295,21 @@ HostIF_LockPage(VMDriver *vm, void *addr)
 		 * unlocked. 
 		 */
 
+		/*
+		 * fubyte forces the page in core
+		 * XXX Have we validated the pointer before? If not, we should check the
+		 *     return value --hpreg
+		 */
 		c = fubyte(addr);
 
 		/*
 		 * now locate it and write the results back to the pmap
 		 *
 		 * Under extreme memory pressure, the page may be gone again.
-		 * Just fail the lock in that case.
-		 * (It was ASSERT_BUG(6339, mpn).)
+		 * Just fail the lock in that case. (It was ASSERT_BUG(6339, mpn).)
 		 * -- edward
 		 */
-		mpn = FindMPN(vpn);
+		mpn = HostifVa2Mpn((VA)addr);
 		if (mpn == 0) {
 			return 0;
 		}
@@ -338,6 +326,19 @@ HostIF_LockPage(VMDriver *vm, void *addr)
 		}
       
 		HOST_LOCK_PFN(vm, mpn);
+
+		if (HostifVa2Mpn((VA)addr) != mpn) {
+			/*
+			 * Our page was swapped out (and possibly swapped in again at another
+			 * location) between the previous HostifVa2Mpn() and HOST_LOCK_PFN().
+			 * So we have just locked the wrong page. Report failure and let
+			 * userland retry --hpreg
+			 */
+
+			HOST_UNLOCK_PFN(vm, mpn);
+
+			return 0;
+		}
 		/*
 		 * If the entry doesn't exist, add it to the memtracker
 		 * otherwise we just update the mpn.
@@ -356,7 +357,7 @@ HostIF_LockPage(VMDriver *vm, void *addr)
 		 * Already locked. 
 		 */
 
-		mpn = FindMPN(vpn);
+		mpn = HostifVa2Mpn((VA)addr);
 		ASSERT(mpn);
 		ASSERT(entryPtr->mpn == mpn);
 		ASSERT(HOST_ISTRACKED_PFN(vm, mpn));
@@ -388,7 +389,7 @@ HostIF_UnlockPage(VMDriver *vm, void *addr)
 	MemTrackEntry *entryPtr;
 
 	vpn = PTR_2_VPN(addr);
-	entryPtr = (MemTrackEntry *) MemTrack_LookupVPN(vm->memtracker, vpn);
+	entryPtr = MemTrack_LookupVPN(vm->memtracker, vpn);
 
 	if ((entryPtr == NULL) || (entryPtr->mpn == 0)) {
 		Warning("HostIF_UnlockPage vpn=0x%06x not tracked!\n", vpn);
@@ -448,6 +449,7 @@ HostIF_FreeAllResources(VMDriver *vm)
 #if 0
 	/* XXXX causes a panic? */
 	if (vm->crossvaddr != NULL) {
+		/* XXX first vsunlock? */
 		uvm_km_free_wakeup(kernel_map, (vaddr_t)vm->crossvaddr,
 		    PAGE_SIZE);
 		uvm_vsunlock(curproc, vm->crossuaddr, PAGE_SIZE);
@@ -577,7 +579,9 @@ HostIF_ReadTime(void)
  */
  
 int     
-HostIF_CopyFromUser(void *dst, void *src, int len)
+HostIF_CopyFromUser(void *dst, // OUT
+		    void *src, // IN
+		    int len)   // IN
 {  
 	return copyin(src, dst, len);
 }
@@ -598,7 +602,9 @@ HostIF_CopyFromUser(void *dst, void *src, int len)
  */   
    
 int
-HostIF_CopyToUser(void *dst, void *src, int len)
+HostIF_CopyToUser(void *dst, // OUT
+		  void *src, // IN
+		  int len)   // IN
 {
 	return copyout(src, dst, len);
 }
@@ -606,25 +612,28 @@ HostIF_CopyToUser(void *dst, void *src, int len)
 /*
  *----------------------------------------------------------------------
  *
- * HostIF_UserToDriverPtr 
+ * HostIF_UserToDriverPtr
  *
- *      Convert user address into an address that can be used by the
- *      driver to access the memory. 
+ *    Convert a user virtual address into a kernel virtual address to access
+ *    the same physical memory address --hpreg
  *
  * Results:
- *      A pointer or NULL on error. 
+ *    The kernel virtual address on success
+ *    NULL on failure
  *
  * Side effects:
+ *    XXX Store the struct page of the crosspage
  *
- * Bugs:
- *      anyone wants to lock the page?
- *----------------------------------------------------------------------
+ *-----------------------------------------------------------------------------
  */
 
+#ifdef DEBUG
 char userpage[NBPG];
+#endif
 
 void *
-HostIF_UserToDriverPtr(VMDriver *vm, void *addr)
+HostIF_UserToDriverPtr(VMDriver *vm, // IN
+		       void *addr)   // IN
 {
 	paddr_t paddr;
 	vaddr_t kvaddr, uaddr;
@@ -693,35 +702,18 @@ HostIF_IsMPNLocked(VMDriver *vm, MPN mpn)
 	entry = MemTrack_Scan(vm->memtracker,vm, FindFunc);
 	return (entry != NULL);
 }
-
-/*
- *----------------------------------------------------------------------
- *
- * HostIF_CheckMemory
- *
- *      Make sure memory locking is OK.
- *
- *      This function runs as part of the application
- *      process, as such FindMPN, get_user and put_user are fine
- *
- * Results:
- *      None
- *
- * Side effects:
- *      None.
- *---------------------------------------------------------------------- 
- */
  
 static void *
-CheckFunc(void *arg, MemTrackEntry *entryPtr)  
+CheckFunc(void *arg,               // IN
+	  MemTrackEntry *entryPtr) // IN
 { 
 	VMDriver *vm = (VMDriver *)arg;
 	MPN mpn;
  
 	if (entryPtr->mpn == 0)
-	return NULL;
+		return NULL;
  
-	mpn = FindMPN(entryPtr->vpn);
+	mpn = HostifVa2Mpn(VPN_2_VA(entryPtr->vpn));
 	if (mpn == 0) {
 		/* unmapped */
 		return entryPtr; 
@@ -736,6 +728,24 @@ CheckFunc(void *arg, MemTrackEntry *entryPtr)
 	return NULL;
 }
 
+/*
+ *-----------------------------------------------------------------------------
+ *
+ * HostIF_CheckMemory
+ *
+ *      Make sure memory locking is OK.
+ *
+ *      This function runs as part of the application process, as such
+ *      HostifVa2Mpn is fine
+ *
+ * Results:
+ *      None
+ *
+ * Side effects:
+ *      None.
+ *
+ *-----------------------------------------------------------------------------
+ */
 VA 
 HostIF_CheckMemory(VMDriver *vm)
 {  
@@ -751,55 +761,6 @@ HostIF_CheckMemory(VMDriver *vm)
 	return 0;
 }
 
-/*
- *----------------------------------------------------------------------
- *
- * HostIF_MhzRating
- *
- *      Return an estimate the of the processor's MHZ rating, based on
- *      the ratio of the cycle counter and gettimeofday
- *      
- * Results: 
- *
- * 
- * Side effects:
- *      None.
- *----------------------------------------------------------------------
- */     
-uint32
-HostIF_MhzRating(VMDriver *vm)
-{
-
-	VmTimeType vDiff, rDiff;
-	uint32 flags;
-
-	/* 
-	 * Compute the cycle rate based on the change in the
-	 * cycle counter and real time clock since we open the
-	 * device.
-	 */
-   
-	SAVE_FLAGS(flags);
-	CLEAR_INTERRUPTS();
-	vDiff = GET_TSC() - vm->startTime.count;
-	rDiff = HostIF_ReadTime() - vm->startTime.time;
-	RESTORE_FLAGS(flags);
-       
-	if (rDiff == 0)
-		return 0;
-
-	/* Need to do 32bit divide since 64bit one doesn't seem to
-	 * work in a linux device driver. Scale the numbers back into
-	 * 32bit numbers. 
-	 * XXX doesn't apply to NetBSD -- redo this one.
-	*/
-	while (vDiff > (0xffffffff/10)) {
-		vDiff >>= 1; 
-		rDiff >>= 1;
-	}
-	return ((10*(uint32)vDiff)/(uint32)rDiff + 5)/10;
-}
-
 /*
  *----------------------------------------------------------------------
  *
@@ -866,10 +827,31 @@ HostIF_GlobalVMLock(int callerID)
 {
 #ifdef SMP_GLOBAL_VMLOCK
 	if (!simple_lock_try(&vmmon_lock)) {
+		/*
+		 * Don't rely too much on this information: imagine the following
+		 * scenario:
+		 *
+		 * process 1               process 2
+		 * ---------               --------
+		 * Acquire vm_lock
+		 * Set vm_lock_holder to 1
+		 * Release vm_lock
+		 * Acquire vm_lock
+		 *                         Try to acquire vm_lock and fail, so report
+		 *                         a bogus vm_lock of 1 instead of the correct 2
+		 * Set vm_lock_holder to 2
+		 * Release vm_lock
+		 *                         Acquire vm_lock
+		 *                         Set vm_lock_holder to 3
+		 *                         Release vm_lock
+		 *
+		 *  --hpreg
+		 */
 		printf("vmmon: %d wait for global VM lock %d\n",
 		    callerID, vm_lock_holder);
 		simple_lock(&vmmon_lock);
 	}
+	vmmon_lock_task = curproc;
 	vmmon_lock_holder = callerID;
 #endif
 }
@@ -895,17 +877,18 @@ void
 HostIF_GlobalVMUnLock(int callerID)
 {
 #ifdef SMP_GLOBAL_VMLOCK
-	if (vm_lock_holder != callerID) {
+	if (vmmon_lock_holder != callerID) {
 		printf("/dev/vmmon: %d releasing global VM lock %d\n",
-		    callerID, vm_lock_holder);
+		    callerID, vmmon_lock_holder);
 	}
-	vm_lock_holder = -1;
-	simple_unlock(&vm_lock); 
+	vmmon_lock_task = NULL;
+	vmmon_lock_holder = -1;
+	simple_unlock(&vmmon_lock); 
 #endif
 }
 
 MA
-HostIF_APIC_Base(VMDriver *vm, Bool setVMPtr, Bool probe)
+HostIF_APICBase(VMDriver *vm, Bool setVMPtr, Bool probe)
 {
 	return 0;
 }
@@ -929,7 +912,475 @@ HostIF_APIC_Base(VMDriver *vm, Bool setVMPtr, Bool probe)
  */
  
 MA      
-HostIF_IOAPIC_Base(VMDriver *vm) 
+HostIF_IOAPICBase(VMDriver *vm) 
 {
 	return 0;
 }
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * GetMSR --
+ *
+ *      Wrapper for the macro that calls the rdmsr instruction.
+ *
+ * Results:
+ *      Returns the value of the MSR.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static INLINE uint64
+GetMSR(int index)
+{
+   uint64 msr;
+   msr = __GET_MSR (index);
+   return msr;
+}
+
+uint64
+HostIF_RDMSR(int index)
+{
+   return GetMSR(index);
+}
+
+void
+HostIF_WRMSR(int index, uint64 val)
+{
+   uint32 hi = val >> 32;
+   uint32 low = val & (-1UL);
+   asm("wrmsr" : : "c" (index), "a" (low), "d" (hi));
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * GetCPUID --
+ *
+ *      Wrapper for the macro that calls the cpuid instruction.
+ *
+ * Results:
+ *      Returns the value of cpuid in regs.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static INLINE void
+GetCPUID(int index, uint32 *regs)
+{
+   __GET_CPUID (index, regs);
+}
+
+void
+HostIF_GetCPUID(int index, uint32 *regs)
+{
+   GetCPUID(index, regs);
+}
+
+
+#ifdef notyet
+
+/*
+ * Utility routines for accessing and enabling the APIC
+ */
+
+#if defined(KERNEL_2_1) || defined(USE_PERFCOUNTERS)
+
+/*
+ * Defines for accessing the APIC.  We use readl/writel to access the APIC
+ * which is how Linux wants you to access I/O memory (though on the x86
+ * just dereferencing a pointer works just fine).
+ */
+#define APICR_TO_ADDR(apic, reg)      (apic + (reg << 4))
+#define GET_APIC_REG(apic, reg)       (readl(APICR_TO_ADDR(apic, reg)))
+#define SET_APIC_REG(apic, reg, val)  (writel(val, APICR_TO_ADDR(apic, reg)))
+
+#define APIC_MAXLVT(apic)             ((GET_APIC_REG(apic, APICR_VERSION) >> 16) & 0xff)
+#define APIC_VERSIONREG(apic)         (GET_APIC_REG(apic, APICR_VERSION) & 0xff)
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * SetVMAPICPtr --
+ *
+ *      Sets a pointer to the APIC's virtual address in the VMDriver
+ *      structure.  
+ *
+ * Results:
+ *      None.
+ *
+ * Side effects:
+ *      The VMDriver structure is updated.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static void
+SetVMAPICPtr(VMDriver *vm, MPN mpn)
+{
+   volatile void *hostapic;
+
+   hostapic = (volatile void *) ioremap_nocache(MPN_2_MA(mpn), PAGE_SIZE);
+   if (hostapic) {
+      if ((APIC_VERSIONREG(hostapic) & 0xF0) == 0x10) {
+	 vm->hostAPIC = (volatile uint32 (*)[4]) hostapic;
+      }
+   }
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * ProbeAPIC --
+ *
+ *      Find the base physical address of the APIC.  On P6 family
+ *      processors, this is done by reading the address from an MSR.
+ *
+ * Results:
+ *      Returns the physical address of the APIC if found, 0 otherwise.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+MA
+ProbeAPIC(VMDriver *vm,       // IN/OUT: driver state
+	  Bool setVMPtr,      // IN: set a pointer to the APIC's virtual address
+	  Bool forceEnable)   // IN: APIC is being enabled here in the driver 
+{
+   uint64 msr;
+   MPN mpn;
+   CPUOem cpuOem = CPUID_GetOEM();
+
+   if (cpuOem == CPUID_OEM_INTEL) {
+      uint32 version = CPUID_GetVersion();
+      uint32 features = CPUID_GetFeatures();
+
+      if ((features & CPUID_FEATURE_COMMON_MSR) &&
+	  (forceEnable || (features & CPUID_FEATURE_COMMON_APIC))) {
+
+	 /* APIC is present and enabled */
+         if (CPUID_FAMILY_IS_P6(version) || CPUID_FAMILY_IS_PENTIUM4(version)) {
+	    /* P6 allows APIC relocation */
+	    msr = GetMSR(APIC_BASE_MSR);
+	    mpn = (msr >> 12) & 0xFFFFFF;
+	    if (setVMPtr) {
+	       /* obtain a system address the APIC, only for P6.
+		  Not recommended for P5, per Intel Book 3, page 7-16 */
+	       SetVMAPICPtr(vm, mpn);
+	    } else {
+	       vm->hostAPIC = NULL;
+	    }
+	    return MPN_2_MA(mpn);
+	 } else if (!forceEnable) {
+	    return APIC_DEFAULT_ADDRESS;
+	 }
+      }
+   }
+   return 0;
+}
+#endif
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * HostIF_APICBase --
+ *
+ *      Lookup the APIC physical address
+ *
+ * Results:
+ *      
+ *      Returned value is APIC physical address
+ *      
+ *
+ * Side effects:
+ *     None
+ *
+ *----------------------------------------------------------------------
+ */
+
+MA
+HostIF_APICBase(VMDriver *vm, Bool setVMPtr, Bool probe)
+{
+#if (defined(KERNEL_2_1) && defined(__SMP__)) || defined(CONFIG_X86_UP_IOAPIC) || defined(CONFIG_X86_UP_APIC)
+   VA kAddr;
+   MA ma;
+
+   if (probe) {
+      ma = ProbeAPIC(vm, setVMPtr, FALSE);
+      if (ma != 0) {
+	 return ma;
+      }
+   }
+
+   kAddr = __fix_to_virt(FIX_APIC_BASE);
+   if (setVMPtr) {
+      vm->hostAPIC = (void *)kAddr;
+   } else {
+      vm->hostAPIC = NULL;
+   }
+
+   return MPN_2_MA(HostifVa2Mpn(kAddr));
+#else
+   return 0;
+#endif
+}
+
+
+#ifdef USE_PERFCOUNTERS
+
+/* 
+ * Defines for masking/unmasking PIC interrupts. 
+ */
+#define PIC_PRIMARY_MASK_PORT         0x21
+#define PIC_SECONDARY_MASK_PORT       0xA1
+
+#define MaskPICIRQs(mask1, mask2)        \
+   mask1 = INB(PIC_PRIMARY_MASK_PORT);   \
+   OUTB(PIC_PRIMARY_MASK_PORT, 0xff);    \
+   mask2 = INB(PIC_SECONDARY_MASK_PORT); \
+   OUTB(PIC_SECONDARY_MASK_PORT, 0xff);     
+
+#define UnmaskPICIRQs(mask1, mask2)      \
+   OUTB(PIC_PRIMARY_MASK_PORT, mask1);   \
+   OUTB(PIC_SECONDARY_MASK_PORT, mask2); 
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * HostIF_APICEnable --
+ *
+ *      Enable the APIC, if not already enabled.  The APIC is put into
+ *      through local mode.  In this mode the APIC passes interrupts 
+ *      along to the processor just as if they had come directly from the PIC.
+ *
+ * Results:
+ *      Returned value is APIC physical address.
+ *      
+ * Side effects:
+ *      APIC gets enabled.      
+ *
+ *----------------------------------------------------------------------
+ */
+
+MA 
+HostIF_APICEnable(VMDriver *vm)
+{
+   MA ma;
+   volatile void *va;
+   volatile uint8 picMask1, picMask2;
+   uint32 apicMSR, dummy;
+   uint32 tempMSR;
+   uint32 flags;
+
+   ASSERT(vm);
+   ASSERT(vm->hostAPIC == 0);
+
+   /*
+    *  Find the APIC's physical address. 
+    */
+   ma = ProbeAPIC(vm, FALSE, TRUE);
+   if (ma == 0) {
+      printk(KERN_WARNING "/dev/vmmon: Unable to find APIC's physaddr\n");
+      return 0;
+   }
+
+   /*
+    * Map APIC's physical address into a virtual address.
+    */
+   va = (volatile void *) ioremap_nocache(ma, PAGE_SIZE);
+   if (va == 0) {
+      printk(KERN_WARNING "/dev/vmmon: Unable to map APIC at physaddr 0x%x\n", ma);
+      return 0;
+   } 
+
+   /*
+    * Mask interrupts and enable the APIC.  Intel's Software Developer's Guide
+    * Guide Vol III, Appendix E says that we need to mask PIC interrupts before
+    * programming the local interrupt pins.
+    */
+   SAVE_FLAGS(flags); 
+   CLEAR_INTERRUPTS();
+   MaskPICIRQs(picMask1, picMask2);
+
+   RDMSR(APIC_BASE_MSR, apicMSR, dummy);
+   tempMSR = apicMSR;
+
+   if (!(apicMSR & APIC_MSR_ENABLED)) {
+      apicMSR |= APIC_MSR_ENABLED;
+
+      /*
+       * The enable bit of the APIC_BASE_MSR must be set before
+       * we can even read the APIC.
+       */
+      WRMSR(APIC_BASE_MSR, apicMSR, dummy);
+   }
+
+   /*
+    * Set up the local vector table entries.
+    */
+   SET_APIC_REG(va, APICR_SVR, APIC_SVR_ONES|APIC_SVR_VECTOR|APIC_SVR_APICENABLE);
+   SET_APIC_REG(va, APICR_LVT0, APIC_VTE_MODE_EXTINT);
+   SET_APIC_REG(va, APICR_LVT1, APIC_VTE_MODE_NMI);
+   SET_APIC_REG(va, APICR_ERRLVT, APIC_VTE_MASK);
+   SET_APIC_REG(va, APICR_TIMERLVT, APIC_VTE_MASK);
+
+   UnmaskPICIRQs(picMask1, picMask2);
+   RESTORE_FLAGS(flags); 
+
+   vm->hostAPIC =  (volatile uint32 (*)[4]) va;
+   return ma;
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * HostIF_APICSetPerfCtr --
+ *
+ *      Set the performance counter entry in the APIC. 
+ *
+ * Results:
+ *      TRUE if successful, FALSE otherwise.
+ *      
+ * Side effects:
+ *      APIC's PCLVT register is written.
+ *
+ *----------------------------------------------------------------------
+ */
+
+Bool
+HostIF_APICSetPerfCtr(VMDriver *vm,   // IN: driver state
+		      int gate)       // IN: interrupt gate for perfctr interrupts
+{
+   volatile void *apic;
+   uint32 maxlvt;
+
+
+   if (vm->hostAPIC == 0) {
+      printk(KERN_WARNING "PerfCtr: Could not set PCLVT, APIC not available\n");
+      return FALSE;
+   }
+
+   apic = (volatile void *) vm->hostAPIC;
+   maxlvt = APIC_MAXLVT(apic);
+
+   /*
+    * We need maximum LVT entries to be >= 4 in order for there to be
+    * an entry for the performance counters.
+    */
+   if (maxlvt >= 4) {
+      SET_APIC_REG(apic, APICR_PCLVT, APIC_VTE_MODE_FIXED | gate);
+   } else {
+      printk(KERN_WARNING "PerfCtr: Could not set PCLVT, APIC maxlvt %d < 4", maxlvt);
+      return FALSE;
+   }
+
+   return TRUE;
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * HostIF_APICAck --
+ *
+ *      Send an ack to the APIC
+ *
+ * Results:
+ *      None.
+ *      
+ * Side effects:
+ *      APIC's EOI register is written.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void 
+HostIF_APICAck(VMDriver *vm)
+{
+   volatile void *apic;
+   volatile uint32 dummy;
+
+   apic = (volatile void *) vm->hostAPIC;
+
+   /*
+    * Send the APIC an EOI.  Dummy read is copied from Linux 
+    * ack_APIC_irq() routine. 
+    */
+   if (apic) {
+      dummy = GET_APIC_REG(apic, APICR_SVR);
+      SET_APIC_REG(apic, APICR_EOI, 0);
+   }
+}
+
+#endif // ifdef USE_PERFCOUNTERS
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * HostIF_IOAPICBase --
+ *
+ *      Lookup the IO-APIC physical address
+ *
+ * Results:
+ *      
+ *      Returned value is IO-APIC physical address
+ *      
+ *
+ * Side effects:
+ *     None
+ *
+ *----------------------------------------------------------------------
+ */
+#define PCI_CONF_ADDR_REG_PORT   0xcf8
+#define PCI_CONF_DATA_REG_PORT   0xcfc
+#define VMWARE__PCI_DEVICE_ID_INTEL_82371SB 0x70008086
+#define VMWARE__PCI_DEVICE_ID_INTEL_82371AB 0x71108086
+
+MA
+HostIF_IOAPICBase(VMDriver *vm)
+{
+#if (defined(KERNEL_2_1) && defined(__SMP__)) || defined(CONFIG_X86_UP_IOAPIC)
+   int bus, slot, confword;
+   uint32 id, apicbase;
+
+   bus = 0;
+   for (slot = 0; slot < 32; slot++) {
+      confword = 0x80000000 | (bus << 16) |  (slot << 11);
+      OUT32(PCI_CONF_ADDR_REG_PORT, confword);
+      id = IN32(PCI_CONF_DATA_REG_PORT);
+      if ((id == VMWARE__PCI_DEVICE_ID_INTEL_82371SB) ||
+	  (id == VMWARE__PCI_DEVICE_ID_INTEL_82371AB)) {
+#ifdef VMX86_DEVEL
+	 Log("IOAPIC: detected %s on PCI bus %d, slot %d\n",
+	     (id == VMWARE__PCI_DEVICE_ID_INTEL_82371SB) ? "PIIX3" : "PIIX4",
+	     bus, slot);
+#endif
+	 confword |= 20; /* register 80 */
+	 OUT32(PCI_CONF_ADDR_REG_PORT, confword);
+	 apicbase = IN32(PCI_CONF_DATA_REG_PORT);
+	 apicbase = apicbase & 0x3f;
+	 return (IOAPIC_DEFAULT_ADDRESS | (apicbase << 10));
+      }
+   }
+   
+   return MPN_2_MA(HostifVa2Mpn(__fix_to_virt(VMWARE__FIX_IO_APIC_BASE)));
+#else
+   return 0;
+#endif
+}
+
+#endif /* notyet */
