@@ -1,5 +1,5 @@
 /*
- * Copyright 2001 Wasabi Systems, Inc.
+ * Copyright (c) 2001 Wasabi Systems, Inc.
  * All rights reserved.
  *
  * Written by Frank van der Linden for Wasabi Systems, Inc.
@@ -87,9 +87,11 @@ static int ether_ioctl(struct ifnet *, u_long, caddr_t);
 #include <vm/vm.h>
 #endif
 
+extern struct vm_map *kernel_map;
+
 #define HUBDEBUG  if (hub_debug) printf
 
-int if_hub_lkmentry(struct lkm_table *, int, int);
+int vmnet_lkmentry(struct lkm_table *, int, int);
 
 static int hub_open(dev_t dev, int oflags, int devtype, struct proc *p);
 static int hub_close(dev_t dev, int cflags, int devtype, struct proc *p);
@@ -110,6 +112,8 @@ static void port_netifattach(struct hubport_softc *, const char *);
 static int port_pktmatch(struct hubport_softc *, struct mbuf *);
 static struct hubport_softc *port_allocate(struct hubdev_softc *);
 static void port_deallocate(struct hubport_softc *);
+static int port_mappoll(struct hubport_softc *, vaddr_t);
+static int port_unmappoll(struct hubport_softc *);
 
 static int hub_sendchain(struct hubport_softc *, struct mbuf *);
 static int hub_fake_clonedev(dev_t, int, struct proc *);
@@ -128,14 +132,10 @@ static int hub_debug = 0;
 static struct hubdev_softc *hub_scs[MAXHUBDEVS];
 
 
-#if __NetBSD_Version__ >= 106080000
-MOD_DEV("vmnet", "vmnet", NULL, -1, &hub_dev, -1)
-#else
 MOD_DEV("vmnet", LM_DT_CHAR, -1, &hub_dev)
-#endif
 
 int
-if_hub_lkmentry(struct lkm_table *lkmtp, int cmd, int ver)
+vmnet_lkmentry(struct lkm_table *lkmtp, int cmd, int ver)
 {
 	DISPATCH(lkmtp, cmd, ver, hub_handle, hub_handle, hub_handle)
 }
@@ -215,6 +215,8 @@ port_deallocate(struct hubport_softc *portsc)
 		if_detach(&portsc->port_if);
 		splx(s);
 	}
+	if (portsc->port_pollptr != NULL)
+		port_unmappoll(portsc);
 	FREE(portsc, M_DEVBUF);
 }
 
@@ -354,6 +356,62 @@ port_pktmatch(struct hubport_softc *portsc, struct mbuf *m)
 	return 0;
 }
 
+static int
+port_mappoll(struct hubport_softc *portsc, vaddr_t uaddr)
+{
+	vaddr_t kaddr, upage, kpage;
+	paddr_t paddr;
+	voff_t off;
+	int error;
+
+	upage = trunc_page(uaddr);
+	off = uaddr - upage;
+	if (off > (PAGE_SIZE - sizeof(u_int32_t)))
+		return EINVAL;
+	kpage = uvm_km_valloc_wait(kernel_map, PAGE_SIZE);
+	kaddr = kpage + off;
+
+	PHOLD(curproc);
+	error = uvm_vslock(curproc, (void *)upage, PAGE_SIZE,
+	    VM_PROT_READ|VM_PROT_WRITE);
+	if (error)
+		goto err;
+
+	pmap_extract(vm_map_pmap(&curproc->p_vmspace->vm_map), upage, &paddr);
+	pmap_kenter_pa(kpage, paddr, VM_PROT_READ|VM_PROT_WRITE);
+
+#if 1
+	pmap_update(pmap_kernel());
+#endif
+	PRELE(curproc);
+	portsc->port_kpollpage = kpage;
+	portsc->port_upollpage = upage;
+	portsc->port_pollptr = (u_int32_t *)kaddr;
+	return 0;
+
+    err:
+	PRELE(curproc);
+	uvm_km_free_wakeup(kernel_map, kpage, PAGE_SIZE);
+	return (error);
+}
+
+static int
+port_unmappoll(struct hubport_softc *portsc)
+{
+	pmap_kremove(portsc->port_kpollpage, PAGE_SIZE);
+#if 1
+	pmap_update(pmap_kernel());
+#endif
+	uvm_km_free_wakeup(kernel_map, portsc->port_kpollpage, PAGE_SIZE);
+	PHOLD(curproc);
+	uvm_vsunlock(curproc, (void *)portsc->port_upollpage, PAGE_SIZE);
+	PRELE(curproc);
+	portsc->port_pollptr = NULL;
+	portsc->port_kpollpage = 0;
+	portsc->port_upollpage = 0;
+	return 0;
+}
+
 
 static int
 hub_open(dev_t dev, int flag, int mode, struct proc *p)
@@ -590,6 +648,9 @@ hub_sendchain(struct hubport_softc *portsc, struct mbuf *m)
 			}
 		}
 		if (sent != 0) {
+			if (psc2->port_pollptr != NULL &&
+			    psc2->port_rcvq.ifq_len >= portsc->port_pollthresh)
+				*psc2->port_pollptr |= portsc->port_pollmask;
 			if (psc2->port_flags & PORT_RWAIT) {
 				psc2->port_flags &= ~PORT_RWAIT;
 				wakeup((caddr_t)psc2);
@@ -667,6 +728,7 @@ hub_ioctl(dev_t dev, u_long cmd, caddr_t data, int flag, struct proc *p)
 	struct timeval tv;
 	int zero = 0;
 	u_long val;
+	u_int32_t upolladdr, pollmask, dummy;
 
 	HUBDEBUG("if_hub: ioctl %lx on hub %d port %d by pid %d\n",
 	    cmd, HUBUNIT(dev), HUBPORT(dev), p->p_pid);
@@ -822,6 +884,40 @@ hub_ioctl(dev_t dev, u_long cmd, caddr_t data, int flag, struct proc *p)
 			memcpy(portsc->port_addr, sma.addr,
 			    sizeof sma.addr);
 			break;
+		case LINUX_VMWARE_SETNOTIFY:
+			if (portsc->port_pollptr != NULL)
+				return EINVAL;
+			error = copyin(pt->data, &upolladdr, 4);
+			if (error != 0)
+				return error;
+			error = copyin((u_int32_t *)pt->data + 1, &pollmask, 4);
+			if (error != 0)
+				return error;
+
+			/* check the supplied address is valid */
+			dummy = 0;
+			error = copyout(&dummy, (void *)upolladdr, 4);
+			if (error != 0)
+				return error;
+
+			/* create kernel mapping for the user address */
+			error = port_mappoll(portsc, (vaddr_t)upolladdr);
+			if (error != 0)
+				return error;
+			portsc->port_pollmask = pollmask;
+			portsc->port_pollthresh = 1;
+			break;
+		case LINUX_VMWARE_UNSETNOTIFY:
+			if (portsc->port_pollptr == NULL)
+				return EINVAL;
+			port_unmappoll(portsc);
+			portsc->port_pollthresh = 0;
+			break;
+		case LINUX_VMWARE_SETCLUSTERSIZE:
+			if (portsc->port_pollptr == NULL)
+				return EINVAL;
+			portsc->port_pollthresh = (unsigned)pt->data;
+			break;
 		default:
 			return ENOTTY;
 		}
@@ -903,6 +999,9 @@ hub_read(dev_t dev, struct uio *uio, int flag)
 		m_freem(m0);
 	}
 
+	if (portsc->port_pollptr != NULL && IFQ_IS_EMPTY(&portsc->port_rcvq))
+		*portsc->port_pollptr &= ~portsc->port_pollmask;
+
 	return (error);
 }
 
